// Copyright © 2016–2017 University of Malta

// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

use error::Error;
use input::SourceSet;
use node;
use std::collections::HashMap;
use tree::{Index, Link, Tree};

type IdentMap = HashMap<String, Index>;
type ScopeMap = HashMap<Index, IdentMap>;

pub fn resolve(sources: &SourceSet,
               tree: &mut Tree,
               root_i: Index)
               -> Result<(), Error> {
    let mut scopes = ScopeMap::new();
    build_map_rec(sources, tree, root_i, &mut scopes, None)?;
    resolve_rec(sources, tree, root_i, &scopes, None)?;
    reorder_scopes(sources, tree, root_i)?;
    Ok(())
}

pub fn make_scope_map(tree: &Tree, index: Index) -> HashMap<&str, Index> {
    let mut map = HashMap::new();
    let (_, defs) = node::get_scope(tree, index).unwrap();
    for def_i in defs.iter(tree) {
        let (name, _, _) = node::get_definition(tree, def_i).unwrap();
        map.insert(&name[..], def_i);
    }
    map
}

fn build_map_rec(sources: &SourceSet,
                 tree: &mut Tree,
                 index: Index,
                 scopes: &mut ScopeMap,
                 mut scope: Option<Index>)
                 -> Result<(), Error> {
    if let Some((ident, _, _)) = node::get_definition(tree, index) {
        let map = scopes.get_mut(&scope.unwrap()).unwrap();
        if map.insert(ident.clone(), index).is_some() {
            return Err(Error::general(&format!("{}: identifier {} already \
                                                used in scope",
                                               tree[index]
                                                   .span()
                                                   .loc_string(sources),
                                               ident)));
        }
    }
    let mut children = tree.children(index);
    while let Some(child_i) = children.next(tree) {
        if node::get_scope(tree, child_i).is_some() {
            if scopes.insert(child_i, IdentMap::new()).is_some() {
                panic!("scope has already been resolved")
            }
            scope = Some(child_i);
        }
        build_map_rec(sources, tree, child_i, scopes, scope)?;
    }
    Ok(())
}

fn resolve_ident(sources: &SourceSet,
                 tree: &Tree,
                 index: Index,
                 scopes: &ScopeMap,
                 mut scope: Index,
                 ident: &str)
                 -> Result<Option<Index>, Error> {
    loop {
        let map = scopes.get(&scope).unwrap();
        if let Some(i) = map.get(ident) {
            return Ok(Some(*i));
        }
        if let (Some(next_outer), _) = node::get_scope(tree, scope).unwrap() {
            scope = next_outer;
            continue;
        }
        return Err(Error::general(&format!("{}: identifier {} not found",
                                           tree[index]
                                               .span()
                                               .loc_string(sources),
                                           ident)));
    }
}

fn resolve_rec(sources: &SourceSet,
               tree: &mut Tree,
               index: Index,
               scopes: &ScopeMap,
               mut scope: Option<Index>)
               -> Result<(), Error> {
    let mut link_to_def = None;
    if let Some(ident) = node::get_ident(tree, index) {
        let s = scope.unwrap();
        link_to_def =
            resolve_ident(sources, tree, index, scopes, s, &ident[..])?;
    } else if let Some(ident) = node::get_outer_ident(tree, index) {
        let s = scope.unwrap();
        let (outer_o, _) = node::get_scope(tree, s).unwrap();
        let os = outer_o.unwrap();
        link_to_def =
            resolve_ident(sources, tree, index, scopes, os, &ident[..])?;
    }
    if let Some(def_i) = link_to_def {
        let mut replace = tree.replacement_link(index).unwrap();
        let span = tree[index].span();
        tree.remove(index);
        let resolved_i = node::resolved_ident(def_i, tree);
        tree[resolved_i].set_span(span);
        replace.link(resolved_i, tree);
    }
    let mut children = tree.children(index);
    while let Some(child_i) = children.next(tree) {
        if node::get_scope(tree, child_i).is_some() {
            scope = Some(child_i);
        }
        resolve_rec(sources, tree, child_i, scopes, scope)?;
    }
    Ok(())
}

fn reorder_scopes(sources: &SourceSet,
                  tree: &mut Tree,
                  root_i: Index)
                  -> Result<(), Error> {
    let mut progress = HashMap::new();
    let mut scope_order = HashMap::new();
    match reorder_scope_rec(tree, root_i, &mut progress, &mut scope_order) {
        Ok(_) => {}
        Err(index) => {
            let ident = match node::get_definition(tree, index) {
                Some((ident, _, _)) => ident,
                _ => unreachable!(),
            };
            return Err(Error::general(&format!("{}: identifier {} is in \
                                                recursive cycle",
                                               tree[index]
                                                   .span()
                                                   .loc_string(sources),
                                               ident)));
        }
    }
    Ok(())
}

// progress.get(&index) is:
//     None if definition at index has not been checked yet,
//     Some(&false) if definiton at index is being checked,
//     Some(&true) if definiton at index has no cycles.
fn reorder_scope_rec(tree: &mut Tree,
                     index: Index,
                     progress: &mut HashMap<Index, bool>,
                     scope_order: &mut HashMap<Index, Vec<Index>>)
                     -> Result<(), Index> {
    let def_scope = match node::get_definition(tree, index) {
        Some((_, scope_i, _)) => {
            match progress.get(&index) {
                Some(&true) => return Ok(()),
                Some(&false) => return Err(index),
                None => {}
            }
            progress.insert(index, false);
            Some(scope_i)
        }
        None => None,
    };
    if let Some(def) = node::get_resolved_ident(tree, index) {
        reorder_scope_rec(tree, def, progress, scope_order)?;
    }
    let mut children = tree.children(index);
    while let Some(child_i) = children.next(tree) {
        reorder_scope_rec(tree, child_i, progress, scope_order)?;
    }
    if let Some(scope) = def_scope {
        progress.insert(index, true);
        scope_order.entry(scope).or_insert_with(Vec::new).push(index);
    }
    if let Some(mut idents) = scope_order.remove(&index) {
        let mut link = Link::child(index);
        for i in idents.drain(..) {
            link.link(i, tree);
        }
    }
    Ok(())
}
