// Copyright © 2016–2017 University of Malta

// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

use core;
use error::Error;
use eval;
use idents;
use input::SourceSet;
use node;
use output;
use parse;
use std::fmt;
use tree::{Index, Link, Tree};

pub struct Package {
    sources: SourceSet,
    tree: Tree,
    root_i: Index,
}

impl Package {
    pub fn new() -> Package {
        let mut tree = Tree::new();
        let core_i = node::scope(None, &mut tree);
        let root_i = node::root(core_i, &mut tree);
        core::init(&mut tree, core_i);
        let scope_i = node::scope(Some(core_i), &mut tree);
        let package_i = node::package(scope_i, &mut tree);
        Link::sibling(core_i).link(package_i, &mut tree);
        Package {
            sources: SourceSet::new(),
            tree: tree,
            root_i: root_i,
        }
    }

    pub fn parse_sources(&mut self, filenames: &[&str]) -> Result<(), Error> {
        let (_, mut packages) = node::get_root(&self.tree, self.root_i)
            .unwrap();
        let package_i = packages.next(&self.tree).unwrap();
        let (scope_i, _) = node::get_package(&self.tree, package_i).unwrap();
        let mut link = Link::sibling(scope_i);
        let existing = self.sources.len();
        self.sources.read(filenames)?;
        for (mut span, _) in self.sources.source_spans().skip(existing) {
            let file_i = parse::parse(&self.sources,
                                      &mut span,
                                      &mut self.tree,
                                      scope_i)?;
            link.link(file_i, &mut self.tree);
        }
        Ok(())
    }

    pub fn resolve_idents(&mut self) -> Result<(), Error> {
        idents::resolve(&self.sources, &mut self.tree, self.root_i)?;
        Ok(())
    }

    pub fn evaluate(&mut self) -> Result<(), Error> {
        eval::evaluate(&self.sources, &mut self.tree, self.root_i)?;
        Ok(())
    }

    pub fn output(&self) -> Result<(), Error> {
        let (_, mut packages) = node::get_root(&self.tree, self.root_i)
            .unwrap();
        let package_i = packages.next(&self.tree).unwrap();
        let (_, files) = node::get_package(&self.tree, package_i).unwrap();
        for file_i in files.iter(&self.tree) {
            let (scope_i, _) = node::get_file(&self.tree, file_i).unwrap();
            let (_, defs) = node::get_scope(&self.tree, scope_i).unwrap();
            for def_i in defs.iter(&self.tree) {
                let (ident, _, item_i) =
                    node::get_definition(&self.tree, def_i).unwrap();
                if let Some((_, outputs_i, _, _)) =
                    node::get_function(&self.tree, item_i) {
                    // write only component with list outputs
                    if node::get_list(&self.tree, outputs_i).is_some() {
                        output::output_component(&self.sources,
                                                 &self.tree,
                                                 ident,
                                                 item_i)?;
                    }
                }
            }
        }
        Ok(())
    }
}

impl fmt::Debug for Package {
    fn fmt(&self, fmt: &mut fmt::Formatter) -> fmt::Result {
        write!(fmt, "Sources:")?;
        for (_, filename) in self.sources.source_spans() {
            write!(fmt, " {}", filename.get(&self.sources))?;
        }
        write!(fmt, "\n")?;
        for node_i in self.tree.children(self.root_i).iter(&self.tree) {
            write!(fmt, "{}", self.tree.display(Some(&self.sources), node_i))?;
        }
        Ok(())
    }
}
