// Copyright © 2016–2017 University of Malta

// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

use error::Error;
use gmp_mpfr::{Assign, Float, Integer, Pow, Special};
use input::{SourceSet, Span};
use node::{self, Selector};
use std::cmp;
use std::collections::HashMap;
use tree::{Index, Link, Tree};

#[derive(Clone, Copy, Debug, PartialEq, PartialOrd)]
pub enum TokenKind {
    WideArrow, // =>
    NarrowArrow, // ->
    Equals, // =
    Comma, // ,
    Semicolon, // ;
    Colon, // :
    PlusColon, // +:
    MinusColon, // -:

    // increasing in precedence
    IsEquals, // ==
    NotEquals, // !=
    Less, // <
    LessEquals, // <=
    Greater, // >
    GreaterEquals, // >=

    And, // &
    Or, // |
    Xor, // ^
    Nand, // !&
    Nor, // !|
    Xnor, // !^
    Cat, // ~

    ShiftLeft, // <<
    ShiftRight, // >>

    Plus, // +
    Minus, // -

    Times, // *
    Division, // /
    Remainder, // %

    Power, // **

    // Other unary operators at same level of precedence: & | ^ + -
    // Note that !& !| !^ do not need unary operators as for example
    // !(&x) would be equal to !&x.
    Not, // !

    Point, // .
    OpenParens, // (
    CloseParens, // )
    OpenSquare, // [
    CloseSquare, // ]
    OpenBrace, // {
    CloseBrace, // }

    KeywordAs,
    KeywordElse,
    KeywordFn,
    KeywordIf,
    KeywordMatch,

    Integer,
    IntegerXz,
    Float,
    ZeroX,
    ZeroZ,
    Word,
    Eof,
    Unknown,
}

struct Token {
    kind: TokenKind,
    span: Span,
}

// first character is known to be _ | a...z | A...Z
fn get_word_bytes(s: &str) -> Result<(TokenKind, usize), (String, usize)> {
    let mut iter = s.char_indices();

    let mut bytes = match iter.next() {
        Some((index, c)) if c == '_' || ('a' <= c && c <= 'z') ||
                            ('A' <= c && c <= 'Z') => index + c.len_utf8(),
        _ => unreachable!(),
    };
    for (index, c) in iter {
        if c == '_' || ('a' <= c && c <= 'z') || ('A' <= c && c <= 'Z') ||
           ('0' <= c && c <= '9') {
            bytes = index + c.len_utf8();
        } else {
            break;
        }
    }
    let kind = match &s[0..bytes] {
        "as" => TokenKind::KeywordAs,
        "else" => TokenKind::KeywordElse,
        "fn" => TokenKind::KeywordFn,
        "if" => TokenKind::KeywordIf,
        "match" => TokenKind::KeywordMatch,
        _ => TokenKind::Word,
    };
    Ok((kind, bytes))
}

// Point, ZeroZ, ZeroX, Integer or Float
fn get_number_bytes(s: &str) -> Result<(TokenKind, usize), (String, usize)> {
    let mut base = 10;
    let mut has_point = false;
    let mut iter = s.char_indices();
    let (mut prev, mut bytes): (char, usize) = match iter.next() {
        Some((index, c)) if c == '.' => {
            has_point = true;
            match iter.next() {
                Some((index, c)) if '0' <= c && c <= '9' => {
                    (c, index + c.len_utf8())
                }
                _ => return Ok((TokenKind::Point, index + c.len_utf8())),
            }
        }
        Some((index0, c0)) if c0 == '0' => {
            if let Some((index1, c1)) = iter.next() {
                match c1 {
                    '0'...'9' | '_' => {}
                    '.' => has_point = true,
                    'b' => base = 2,
                    'q' => base = 4,
                    'o' => base = 8,
                    'x' => base = 16,
                    'z' => {
                        return Ok((TokenKind::ZeroZ, index1 + c1.len_utf8()))
                    }
                    _ => {
                        return Ok((TokenKind::Integer, index0 + c0.len_utf8()))
                    }
                }
                (c1, index1 + c1.len_utf8())
            } else {
                return Ok((TokenKind::Integer, index0 + c0.len_utf8()));
            }
        }
        Some((index, c)) if '1' <= c && c <= '9' => (c, index + c.len_utf8()),
        _ => unreachable!(),
    };
    let mut has_digits = base == 10;
    let mut has_mode = false;
    let mut has_exponent = false;
    while let Some((index, c)) = iter.next() {
        let prev_bytes = bytes;
        bytes = index + c.len_utf8();
        match c {
            '_' if prev != '.' => {}
            '0'...'9' if c as i32 - '0' as i32 >= base => {
                return Err((format!("digit '{}' too large for base {}",
                                    c,
                                    base),
                            bytes))
            }
            '0'...'9' => has_digits = true,
            'a'...'f' | 'A'...'F' if base == 16 => has_digits = true,
            'x' | 'z' if base != 10 && !has_point => {
                has_digits = true;
                has_mode = true;
            }
            '.' if prev != '_' && !has_mode && !has_point => {
                has_point = true;
            }
            'e' | 'E' if base == 10 => {
                has_exponent = true;
                break;
            }
            'p' | 'P' if base != 10 && !has_mode => {
                if !has_digits {
                    return Err(("cannot have floating-point literal with no \
                                 digits"
                                    .into(),
                                bytes));
                }
                has_exponent = true;
                break;
            }
            _ => {
                bytes = prev_bytes;
                break;
            }
        }
        prev = c;
    }
    if has_exponent {
        let (index, c) = match iter.next() {
            Some((index, c)) if c == '+' || c == '-' => (index, c),
            Some((index, c)) if '0' <= c && c <= '9' => (index, c),
            _ => {
                return Err(("exponent must have at least one decimal digit"
                                .into(),
                            bytes))
            }
        };
        has_digits = '0' <= c && c <= '9';
        bytes = index + c.len_utf8();
        for (index, c) in iter {
            if '0' <= c && c <= '9' {
                bytes = index + c.len_utf8();
                has_digits = true;
            } else if !has_digits {
                return Err(("exponent must have  at least one decimal digit"
                                .into(),
                            index));
            } else if c == '_' {
                bytes = index + c.len_utf8();
            } else {
                break;
            }
        }
    }
    if has_exponent || has_point {
        Ok((TokenKind::Float, bytes))
    } else if has_mode {
        Ok((TokenKind::IntegerXz, bytes))
    } else if has_digits {
        Ok((TokenKind::Integer, bytes))
    } else if base == 16 && prev == 'x' {
        // 0x is valid without digits, but cannot have trailing underscores
        Ok((TokenKind::ZeroX, bytes))
    } else {
        Err(("number literal must have at least one digit".into(), bytes))
    }
}

impl TokenKind {
    pub fn to_str(&self) -> &'static str {
        use self::TokenKind::*;
        match *self {
            WideArrow => "=>",
            NarrowArrow => "->",
            Equals => "=",
            Comma => ",",
            Semicolon => ";",
            Colon => ":",
            PlusColon => "+:",
            MinusColon => "-:",
            IsEquals => "==",
            NotEquals => "!=",
            Less => "<",
            LessEquals => "<=",
            Greater => ">",
            GreaterEquals => ">=",
            And => "&",
            Or => "|",
            Xor => "^",
            Nand => "!&",
            Nor => "!|",
            Xnor => "!^",
            Cat => "~",
            ShiftLeft => "<<",
            ShiftRight => ">>",
            Plus => "+",
            Minus => "-",
            Times => "*",
            Division => "/",
            Remainder => "%",
            Power => "**",
            Not => "!",
            Point => ".",
            OpenParens => "(",
            CloseParens => ")",
            OpenSquare => "[",
            CloseSquare => "]",
            OpenBrace => "{",
            CloseBrace => "}",
            KeywordAs => "as",
            KeywordElse => "else",
            KeywordFn => "fn",
            KeywordIf => "if",
            KeywordMatch => "match",
            Integer => "Integer",
            IntegerXz => "IntegerXz",
            Float => "Float",
            ZeroX => "ZeroX",
            ZeroZ => "ZeroZ",
            Word => "Word",
            Eof => "Eof",
            Unknown => "Unknown",
        }
    }

    pub fn reverse_rel(&self) -> TokenKind {
        use self::TokenKind::*;
        match *self {
            Less => Greater,
            LessEquals => GreaterEquals,
            Greater => Less,
            GreaterEquals => LessEquals,
            _ => *self,
        }
    }
}

impl Token {
    fn next(sources: &SourceSet, span: &mut Span) -> Result<Token, Error> {
        span.skip_whitespace(sources);
        let stream = span.get(sources);
        let mut ret = Token {
            kind: TokenKind::Unknown,
            span: *span,
        };
        let mut iter = stream.char_indices();
        let (first_c, first_len) = match iter.next() {
            Some((index, c)) => (c, index + c.len_utf8()),
            None => {
                ret.kind = TokenKind::Eof;
                return Ok(ret);
            }
        };
        ret.span.set_len(first_len as u32);

        // check for two-character symbols
        if let Some((index, second_c)) = iter.next() {
            ret.kind = match (first_c, second_c) {
                ('=', '>') => TokenKind::WideArrow,
                ('-', '>') => TokenKind::NarrowArrow,
                ('+', ':') => TokenKind::PlusColon,
                ('-', ':') => TokenKind::MinusColon,
                ('=', '=') => TokenKind::IsEquals,
                ('!', '=') => TokenKind::NotEquals,
                ('<', '=') => TokenKind::LessEquals,
                ('>', '=') => TokenKind::GreaterEquals,
                ('!', '&') => TokenKind::Nand,
                ('!', '|') => TokenKind::Nor,
                ('!', '^') => TokenKind::Xnor,
                ('<', '<') => TokenKind::ShiftLeft,
                ('>', '>') => TokenKind::ShiftRight,
                ('*', '*') => TokenKind::Power,
                _ => TokenKind::Unknown,
            };
            if ret.kind != TokenKind::Unknown {
                span.advance_start((index + second_c.len_utf8()) as u32);
                ret.span.set_len((index + second_c.len_utf8()) as u32);
                return Ok(ret);
            }
        }

        ret.kind = match first_c {
            '=' => TokenKind::Equals,
            ',' => TokenKind::Comma,
            ';' => TokenKind::Semicolon,
            ':' => TokenKind::Colon,
            '<' => TokenKind::Less,
            '>' => TokenKind::Greater,
            '&' => TokenKind::And,
            '|' => TokenKind::Or,
            '^' => TokenKind::Xor,
            '~' => TokenKind::Cat,
            '+' => TokenKind::Plus,
            '-' => TokenKind::Minus,
            '*' => TokenKind::Times,
            '/' => TokenKind::Division,
            '%' => TokenKind::Remainder,
            '!' => TokenKind::Not,
            '(' => TokenKind::OpenParens,
            ')' => TokenKind::CloseParens,
            '[' => TokenKind::OpenSquare,
            ']' => TokenKind::CloseSquare,
            '{' => TokenKind::OpenBrace,
            '}' => TokenKind::CloseBrace,
            '0'...'9' | '.' => {
                let (kind, bytes) = match get_number_bytes(stream) {
                    Ok(x) => x,
                    Err((msg, bytes)) => {
                        let mut err_span = *span;
                        err_span.set_len(bytes as u32);
                        return Err(err_span.err(sources, &msg));
                    }
                };
                ret.span.set_len(bytes as u32);
                kind
            }
            '_' | 'a'...'z' | 'A'...'Z' => {
                let (kind, bytes) = match get_word_bytes(stream) {
                    Ok(x) => x,
                    Err((msg, bytes)) => {
                        let mut err_span = *span;
                        err_span.set_len(bytes as u32);
                        return Err(err_span.err(sources, &msg));
                    }
                };
                ret.span.set_len(bytes as u32);
                kind
            }
            _ => TokenKind::Unknown,
        };
        span.advance_start(ret.span.len());
        Ok(ret)
    }

    // this must not be called on anything except TokenKind::Integer
    // and TokenKind::IntegerXz
    fn integer_literal(&self, sources: &SourceSet) -> (Integer, Integer) {
        let mut base = 10;
        let mut iter = self.span.get(sources).chars();
        let mut val_buf = iter.next().unwrap() as u32 - '0' as u32;
        if val_buf == 0 {
            if let Some(c) = iter.next() {
                match c {
                    '0'...'9' => val_buf = c as u32 - '0' as u32,
                    '_' => {}
                    'b' => base = 2,
                    'q' => base = 4,
                    'o' => base = 8,
                    'x' => base = 16,
                    _ => unreachable!(),
                };
            } else {
                return (0.into(), 0.into());
            }
        }
        let mut mode_buf = 0u32;
        let (mut val, mut mode) = (Integer::new(), Integer::new());
        let mut counter = 1u32;
        for c in iter {
            let (cval, cmode) = match c {
                '_' => continue,
                '0'...'9' => (c as u32 - '0' as u32, 0),
                'a'...'f' => (c as u32 - 'a' as u32 + 10, 0),
                'A'...'F' => (c as u32 - 'A' as u32 + 10, 0),
                'x' => (0, base - 1),
                'z' => (base - 1, base - 1),
                _ => unreachable!(),
            };
            val_buf = val_buf * base + cval;
            mode_buf = mode_buf * base + cmode;
            counter += 1;
            match (base, counter) {
                (2, 32) | (4, 16) | (16, 8) => {
                    val = (val << 32) + val_buf;
                    mode = (mode << 32) + mode_buf;
                    val_buf = 0;
                    mode_buf = 0;
                    counter = 0;
                }
                (8, 10) => {
                    val = (val << 30) + val_buf;
                    mode = (mode << 30) + mode_buf;
                    val_buf = 0;
                    mode_buf = 0;
                    counter = 0;
                }
                (10, 9) => {
                    val = val * 1000_000_000u32 + val_buf;
                    val_buf = 0;
                    counter = 0;
                }
                _ => {}
            };
        }
        val = match base {
            2 => (val << counter) + val_buf,
            4 => (val << (counter * 2)) + val_buf,
            8 => (val << (counter * 3)) + val_buf,
            10 => val * 10u32.pow(counter) + val_buf,
            16 => (val << (counter * 4)) + val_buf,
            _ => unreachable!(),
        };
        mode = match base {
            2 => (mode << counter) + mode_buf,
            4 => (mode << (counter * 2)) + mode_buf,
            8 => (mode << (counter * 3)) + mode_buf,
            10 => mode,
            16 => (mode << (counter * 4)) + mode_buf,
            _ => unreachable!(),
        };
        (val, mode)
    }

    // this must not be called on anything except TokenKind::Float
    fn float_literal(&self, sources: &SourceSet) -> Float {
        let mut base = 10;
        let mut has_point = false;
        let mut sig_buf = 0;
        let mut digits_after_point = 0;
        let mut iter = self.span.get(sources).chars();
        match iter.next().unwrap() {
            '.' => has_point = true,
            '0' => {
                match iter.next() {
                    Some(c) => {
                        match c {
                            '_' => {}
                            '0'...'9' => sig_buf = c as u32 - '0' as u32,
                            '.' => has_point = true,
                            'b' => base = 2,
                            'q' => base = 4,
                            'o' => base = 8,
                            'x' => base = 16,
                            _ => unreachable!(),
                        }
                    }
                    None => return Float::from((0, 53)),
                }
            }
            c @ '1'...'9' => sig_buf = c as u32 - '0' as u32,
            _ => unreachable!(),
        }
        let mut significant_digits = if sig_buf == 0 { 0 } else { 1 };
        let mut significand = Integer::new();
        let mut counter = 1u32;
        while let Some(c) = iter.next() {
            let cval = match c {
                '_' => continue,
                '.' => {
                    has_point = true;
                    continue;
                }
                '0'...'9' => c as u32 - '0' as u32,
                'a'...'f' if base == 16 => c as u32 - 'a' as u32 + 10,
                'A'...'F' if base == 16 => c as u32 - 'A' as u32 + 10,
                'e' | 'E' | 'p' | 'P' => break,
                _ => unreachable!(),
            };
            if cval != 0 || significant_digits != 0 {
                significant_digits += 1;
            }
            if has_point {
                digits_after_point += 1;
            }
            sig_buf = sig_buf * base + cval;
            counter += 1;
            match (base, counter) {
                (2, 32) | (4, 16) | (16, 8) => {
                    significand = (significand << 32) + sig_buf;
                    sig_buf = 0;
                    counter = 0;
                }
                (8, 10) => {
                    significand = (significand << 30) + sig_buf;
                    sig_buf = 0;
                    counter = 0;
                }
                (10, 9) => {
                    significand = significand * 1000_000_000_u32 + sig_buf;
                    sig_buf = 0;
                    counter = 0;
                }
                _ => {}
            };
        }
        significand = match base {
            2 => (significand << counter) + sig_buf,
            4 => (significand << (counter * 2)) + sig_buf,
            8 => (significand << (counter * 3)) + sig_buf,
            10 => significand * 10_u32.pow(counter) + sig_buf,
            16 => (significand << (counter * 4)) + sig_buf,
            _ => unreachable!(),
        };
        let mut exp_sign = false;
        let mut exp = Some(0i32);
        for c in iter {
            match c {
                '+' | '_' => {}
                '-' => exp_sign = true,
                '0'...'9' => {
                    let digit = c as i32 - '0' as i32;
                    exp = exp.and_then(|x| x.checked_mul(10))
                        .and_then(|x| x.checked_add(digit));
                }
                _ => unreachable!(),
            }
        }
        if exp_sign {
            exp = exp.and_then(i32::checked_neg);
        }
        let mut precision = match base {
            2 => significant_digits,
            4 => significant_digits * 2,
            8 => significant_digits * 3,
            10 => (significant_digits * 10 + 2) / 3,
            16 => significant_digits * 4,
            _ => unreachable!(),
        };
        precision = cmp::max(precision, 53);
        let mut ret = Float::from((significand, precision));
        let exp_per_digit = match base {
            2 | 10 => 1,
            4 => 2,
            8 => 3,
            16 => 4,
            _ => unreachable!(),
        };
        exp =
            exp.and_then(|x| x.checked_sub(digits_after_point * exp_per_digit));
        match exp {
            Some(exp) if base != 10 => ret <<= exp,
            Some(exp) => ret *= Float::from((10, precision)).pow(exp),
            None if exp_sign => ret.assign(Special::Zero),
            None => ret.assign(Special::Infinity),
        }
        ret
    }
}

pub fn parse(sources: &SourceSet,
             span: &mut Span,
             tree: &mut Tree,
             outer_scope_i: Index)
             -> Result<Index, Error> {
    let scope_i = node::scope(Some(outer_scope_i), tree);
    let file_i = node::file(scope_i, tree);
    tree[file_i].set_span(*span);
    let mut link = Link::sibling(scope_i);
    loop {
        span.skip_whitespace(sources);
        if span.is_empty() {
            return Ok(file_i);
        }
        let stat_i = parse_stat(sources, span, tree, scope_i)?;
        if node::get_definition(tree, stat_i).is_none() {
            link.link(stat_i, tree);
        }
    }
}

fn parse_stat(sources: &SourceSet,
              span: &mut Span,
              tree: &mut Tree,
              scope_i: Index)
              -> Result<Index, Error> {
    let orig_span = *span;
    let mut token = Token::next(sources, span)?;
    match token.kind {
        TokenKind::KeywordIf | TokenKind::KeywordMatch => {
            return parse_cond(token, sources, span, tree, scope_i)
        }
        TokenKind::KeywordFn => {
            return parse_func(token, sources, span, tree, scope_i)
        }
        _ => {}
    }
    let point_token = if token.kind == TokenKind::Point {
        let p = token;
        token = Token::next(sources, span)?;
        Some(p)
    } else {
        None
    };
    if token.kind == TokenKind::Word {
        let mut peek = *span;
        if parse_token(sources, &mut peek, TokenKind::Equals, "expected '='")
            .is_ok() {
            *span = peek;
            let ident = token.span.get(sources).to_string();
            let val_i = parse_expr(sources, span, tree, scope_i)?;
            let semicolon = parse_token(sources,
                                        span,
                                        TokenKind::Semicolon,
                                        "expected ';'")?;
            let ass_i = if let Some(p) = point_token {
                let inside_ident = format!("$outer_{}", val_i);
                let span = p.span + semicolon.span;
                let inside_ass_i =
                    node::definition(inside_ident, scope_i, val_i, tree);
                tree[inside_ass_i].set_span(span);
                Link::child(scope_i).link(inside_ass_i, tree);
                let inwards_i = node::resolved_ident(inside_ass_i, tree);
                tree[inwards_i].set_span(span);
                let (outer_o, _) = node::get_scope(tree, scope_i).unwrap();
                let outer_scope_i = outer_o.unwrap();
                let outside_ass_i =
                    node::definition(ident, outer_scope_i, inwards_i, tree);
                tree[outside_ass_i].set_span(span);
                Link::child(outer_scope_i).link(outside_ass_i, tree);
                inside_ass_i
            } else {
                let span = token.span + semicolon.span;
                let ass_i = node::definition(ident, scope_i, val_i, tree);
                tree[ass_i].set_span(span);
                Link::child(scope_i).link(ass_i, tree);
                ass_i
            };
            return Ok(ass_i);
        }
    } else if point_token.is_some() {
        return Err(token.span.err(sources, "expected identifier"));
    }
    *span = orig_span;
    if token.kind == TokenKind::OpenBrace {
        return parse_block(sources, span, tree, Some(scope_i));
    }
    let expr_i = parse_expr(sources, span, tree, scope_i)?;
    let stat_i = node::stat_expr(expr_i, tree);
    let semicolon =
        parse_token(sources, span, TokenKind::Semicolon, "expected ';'")?;
    tree[stat_i].set_span(token.span + semicolon.span);
    Ok(stat_i)
}

fn parse_block(sources: &SourceSet,
               span: &mut Span,
               tree: &mut Tree,
               outer_scope: Option<Index>)
               -> Result<Index, Error> {
    let open =
        parse_token(sources, span, TokenKind::OpenBrace, "expected '{'")?;
    let scope_i = node::scope(outer_scope, tree);
    let block_i = node::stat_block(scope_i, tree);
    let mut link = Link::sibling(scope_i);
    loop {
        let mut peek = *span;
        if let Ok(close) = parse_token(sources,
                                       &mut peek,
                                       TokenKind::CloseBrace,
                                       "expected '}'") {
            *span = peek;
            tree[block_i].set_span(open.span + close.span);
            return Ok(block_i);
        }
        let stat_i = parse_stat(sources, span, tree, scope_i)?;
        if node::get_definition(tree, stat_i).is_none() {
            link.link(stat_i, tree);
        }
    }
}

fn parse_cond(mut token: Token,
              sources: &SourceSet,
              span: &mut Span,
              tree: &mut Tree,
              scope_i: Index)
              -> Result<Index, Error> {
    let mut cond_span = token.span;
    let selector_list_i = node::list(tree);
    let mut selector_link = Link::child(selector_list_i);
    loop {
        let selector_i = match token.kind {
            TokenKind::KeywordIf => {
                parse_if(token, sources, span, tree, scope_i)?
            }
            TokenKind::KeywordMatch => {
                parse_match(token, sources, span, tree, scope_i)?
            }
            _ => unreachable!(),
        };
        selector_link.link(selector_i, tree);
        cond_span += tree[selector_i].span();
        let mut peek = *span;
        if parse_token(sources, &mut peek, TokenKind::KeywordElse, "")
            .is_err() {
            break;
        }
        *span = peek;
        token = Token::next(sources, &mut peek)?;
        match token.kind {
            TokenKind::KeywordIf | TokenKind::KeywordMatch => *span = peek,
            TokenKind::OpenBrace => {
                let else_scope_i = node::scope(Some(scope_i), tree);
                let else_stat_i =
                    parse_block(sources, span, tree, Some(else_scope_i))?;
                let else_span = tree[else_stat_i].span();
                let else_selector_i =
                    node::selector_else(else_scope_i, else_stat_i, tree);
                tree[else_selector_i].set_span(else_span);
                selector_link.link(else_selector_i, tree);
                cond_span += else_span;
                break;
            }
            _ => {
                return Err(token.span
                    .err(sources, "expected 'if', 'match' or '{'"))
            }
        }
    }
    let cond_i = node::stat_cond(selector_list_i, tree);
    tree[cond_i].set_span(cond_span);
    build_cond_expressions(sources, tree, cond_i)?;
    Ok(cond_i)
}

enum CondUndef {
    If(Index),
    Match(Index, Index),
    Else,
}

impl CondUndef {
    fn define(&self, def: Index, tree: &mut Tree) -> Index {
        match *self {
            CondUndef::If(cond) => node::cond_if(cond, def, tree),
            CondUndef::Match(sel, pat) => node::cond_match(sel, pat, def, tree),
            CondUndef::Else => node::cond_else(def, tree),
        }
    }
}

fn build_cond_expressions(sources: &SourceSet,
                          tree: &mut Tree,
                          cond_i: Index)
                          -> Result<(), Error> {
    let mut selectors = node::get_stat_cond(tree, cond_i).unwrap();
    let mut map = None;
    while let Some(selector_i) = selectors.next(tree) {
        match node::get_selector(tree, selector_i).unwrap() {
            Selector::If(cond, scope, _) => {
                let cond_def = node::get_resolved_ident(tree, cond).unwrap();
                let undef = CondUndef::If(cond_def);
                build_cond_selector(sources, tree, &mut map, undef, scope)?;
            }
            Selector::Match(sel, mut match_list) => {
                while let Some(group_i) = match_list.next(tree) {
                    let (mut pats, scope, _) =
                        node::get_match_group(tree, group_i).unwrap();
                    while let Some(pat) = pats.next(tree) {
                        let sel_def = node::get_resolved_ident(tree, sel)
                            .unwrap();
                        let pat_def = node::get_resolved_ident(tree, pat)
                            .unwrap();
                        let undef = CondUndef::Match(sel_def, pat_def);
                        build_cond_selector(sources,
                                            tree,
                                            &mut map,
                                            undef,
                                            scope)?;
                    }
                }
            }
            Selector::Else(scope, _) => {
                let undef = CondUndef::Else;
                build_cond_selector(sources, tree, &mut map, undef, scope)?;
            }
        }
    }
    Ok(())
}

struct Evaluating {
    link: Link,
    is_handled: bool,
}

fn build_cond_selector(sources: &SourceSet,
                       tree: &mut Tree,
                       map: &mut Option<HashMap<String, Evaluating>>,
                       undef: CondUndef,
                       scope: Index)
                       -> Result<(), Error> {
    if let Some(ref mut map) = *map {
        build_cond_selector_rest(sources, tree, map, undef, scope)
    } else {
        let mut new_map = HashMap::new();
        build_cond_selector_first(sources, tree, &mut new_map, undef, scope)?;
        *map = Some(new_map);
        Ok(())
    }
}

fn build_cond_selector_first(sources: &SourceSet,
                             tree: &mut Tree,
                             map: &mut HashMap<String, Evaluating>,
                             undef: CondUndef,
                             scope: Index)
                             -> Result<(), Error> {
    let (outer, mut defs) = node::get_scope(tree, scope).unwrap();
    let outer = outer.unwrap();
    while let Some(def_i) = defs.next(tree) {
        let name = {
            let (name, _, _) = node::get_definition(tree, def_i).unwrap();
            name.clone()
        };
        let list_i = node::list(tree);
        let expr_i = node::expr_cond(list_i, tree);
        let new_def_i = node::definition(name.clone(), outer, expr_i, tree);
        Link::child(outer).link(new_def_i, tree);
        let mut list_link = Link::child(list_i);
        let item_i = undef.define(def_i, tree);
        let span = tree[def_i].span();
        tree[item_i].set_span(span);
        list_link.link(item_i, tree);
        let evaluating = Evaluating {
            link: list_link,
            is_handled: false,
        };
        if map.contains_key(&name) {
            gen_err!(tree[def_i],
                     sources,
                     "identifier '{}' already used in conditional",
                     name)?;
        }
        map.insert(name, evaluating);
    }
    Ok(())
}

fn build_cond_selector_rest(sources: &SourceSet,
                            tree: &mut Tree,
                            map: &mut HashMap<String, Evaluating>,
                            undef: CondUndef,
                            scope: Index)
                            -> Result<(), Error> {
    let (_, mut defs) = node::get_scope(tree, scope).unwrap();
    while let Some(def_i) = defs.next(tree) {
        let evaluating = {
            let (name, _, _) = node::get_definition(tree, def_i).unwrap();
            match map.get_mut(name) {
                Some(evaluating) => {
                    if evaluating.is_handled {
                        gen_err!(tree[def_i],
                                 sources,
                                 "identifier '{}' already used in \
                                  conditional",
                                 name)?;
                    }
                    evaluating
                }
                None => {
                    gen_err!(tree[def_i],
                             sources,
                             "identifier '{}' not in all conditional branches",
                             name)?
                }
            }
        };
        let item_i = undef.define(def_i, tree);
        let span = tree[def_i].span();
        tree[item_i].set_span(span);
        evaluating.link.link(item_i, tree);
        evaluating.is_handled = true;
    }
    for (name, evaluating) in map.iter_mut() {
        if !evaluating.is_handled {
            let sibling = evaluating.link.get_sibling().unwrap();
            gen_err!(tree[sibling],
                     sources,
                     "identifier '{}' not in all conditional branches",
                     name)?;
        }
        evaluating.is_handled = false;
    }
    Ok(())
}

fn make_def_ident(name: String,
                  scope_i: Index,
                  index: Index,
                  tree: &mut Tree)
                  -> Index {
    let span = tree[index].span();
    let def_i = node::definition(name, scope_i, index, tree);
    tree[def_i].set_span(span);
    Link::child(scope_i).link(def_i, tree);
    let ident_i = node::resolved_ident(def_i, tree);
    tree[ident_i].set_span(span);
    ident_i
}

fn parse_if(if_token: Token,
            sources: &SourceSet,
            span: &mut Span,
            tree: &mut Tree,
            scope_i: Index)
            -> Result<Index, Error> {
    let condition_i = parse_expr(sources, span, tree, scope_i)?;
    let condition_name = format!("$mux_if_{}", condition_i);
    let condition_ident_i =
        make_def_ident(condition_name, scope_i, condition_i, tree);
    let inner_scope_i = node::scope(Some(scope_i), tree);
    let stat_i = parse_block(sources, span, tree, Some(inner_scope_i))?;
    let stat_span = tree[stat_i].span();
    let selector_i =
        node::selector_if(condition_ident_i, inner_scope_i, stat_i, tree);
    tree[selector_i].set_span(if_token.span + stat_span);
    Ok(selector_i)
}

fn parse_match(match_token: Token,
               sources: &SourceSet,
               span: &mut Span,
               tree: &mut Tree,
               scope_i: Index)
               -> Result<Index, Error> {
    let select_i = parse_expr(sources, span, tree, scope_i)?;
    let select_name = format!("$mux_match_{}", select_i);
    let select_ident_i = make_def_ident(select_name, scope_i, select_i, tree);
    let open =
        parse_token(sources, span, TokenKind::OpenBrace, "expected '{'")?;
    let match_list_i = node::list(tree);
    let mut match_link = Link::child(match_list_i);
    loop {
        let pattern_list_i = node::list(tree);
        let mut pattern_list_span = Default::default();
        let mut pattern_link = Link::child(pattern_list_i);
        loop {
            let pattern_i = parse_expr(sources, span, tree, scope_i)?;
            let pattern_name = format!("$mux_pattern_{}", pattern_i);
            let pattern_ident_i =
                make_def_ident(pattern_name, scope_i, pattern_i, tree);
            pattern_link.link(pattern_ident_i, tree);
            pattern_list_span += tree[pattern_i].span();
            let token = Token::next(sources, span)?;
            match token.kind {
                TokenKind::Comma => {}
                TokenKind::WideArrow => break,
                _ => {
                    return Err(token.span.err(sources, "expected '=>' or ','"))
                }
            }
        }
        tree[pattern_list_i].set_span(pattern_list_span);
        let inner_scope_i = node::scope(Some(scope_i), tree);
        let stat_i = parse_block(sources, span, tree, Some(inner_scope_i))?;
        let stat_span = tree[stat_i].span();
        let match_group_i =
            node::match_group(pattern_list_i, inner_scope_i, stat_i, tree);
        tree[match_group_i].set_span(pattern_list_span + stat_span);
        match_link.link(match_group_i, tree);
        let mut peek = *span;
        if let Ok(close) = parse_token(sources,
                                       &mut peek,
                                       TokenKind::CloseBrace,
                                       "") {
            *span = peek;
            tree[match_list_i].set_span(open.span + close.span);
            let selector_i =
                node::selector_match(select_ident_i, match_list_i, tree);
            tree[selector_i].set_span(match_token.span + close.span);
            return Ok(selector_i);
        }
    }
}

fn parse_func(func_token: Token,
              sources: &SourceSet,
              span: &mut Span,
              tree: &mut Tree,
              scope_i: Index)
              -> Result<Index, Error> {
    let token = Token::next(sources, span)?;
    if token.kind != TokenKind::Word {
        return Err(token.span.err(sources, "expected function identifier"));
    }
    let ty_i = node::function_type(tree);
    let outer_scope_i = get_function_outer_scope(tree, scope_i);
    let port_scope_i = node::scope(Some(outer_scope_i), tree);
    let inputs_i = parse_params(sources, span, tree, scope_i)?;
    parse_token(sources, span, TokenKind::NarrowArrow, "expected '->'")?;
    let outputs_i = parse_params(sources, span, tree, scope_i)?;
    let body_i = parse_block(sources, span, tree, Some(port_scope_i))?;
    let func_i =
        node::function(inputs_i, outputs_i, port_scope_i, body_i, tree);
    let ident = token.span.get(sources).to_string();
    let tv_i = node::typed_value(ty_i, func_i, tree);
    let def_i = node::definition(ident, scope_i, tv_i, tree);
    let def_span = func_token.span + tree[body_i].span();
    tree[def_i].set_span(def_span);
    Link::child(scope_i).link(def_i, tree);
    Ok(def_i)
}

fn get_function_outer_scope(tree: &mut Tree, scope_i: Index) -> Index {
    let mut outer_i = scope_i;
    loop {
        let parent_i = tree.parent(outer_i).unwrap();
        if node::get_file(tree, parent_i).is_some() {
            return outer_i;
        }
        let (outer_o, _) = node::get_scope(tree, outer_i).unwrap();
        outer_i = outer_o.unwrap();
    }
}

fn parse_params(sources: &SourceSet,
                span: &mut Span,
                tree: &mut Tree,
                scope_i: Index)
                -> Result<Index, Error> {
    let open =
        parse_token(sources, span, TokenKind::OpenParens, "expected '('")?;
    let list_i = node::list(tree);
    let mut link = Link::child(list_i);
    let mut need_comma = false;
    loop {
        let mut token = Token::next(sources, span)?;
        if need_comma {
            match token.kind {
                TokenKind::Comma => token = Token::next(sources, span)?,
                TokenKind::CloseParens => {}
                _ => return Err(token.span.err(sources, "expected ')' or ','")),
            }
        }
        let ident = match token.kind {
            TokenKind::Word => token.span.get(sources).to_string(),
            TokenKind::CloseParens => {
                tree[list_i].set_span(open.span + token.span);
                return Ok(list_i);
            }
            _ => {
                return Err(token.span.err(sources, "expected ')' or parameter"))
            }
        };
        parse_token(sources, span, TokenKind::Colon, "expected ':'")?;
        let ty_i = parse_expr(sources, span, tree, scope_i)?;
        let field_i = node::field_def(ident, ty_i, tree);
        link.link(field_i, tree);
        let field_span = token.span + tree[ty_i].span();
        tree[field_i].set_span(field_span);
        need_comma = true;
    }
}

fn parse_expr(sources: &SourceSet,
              span: &mut Span,
              tree: &mut Tree,
              scope_i: Index)
              -> Result<Index, Error> {
    parse_expr_relational(sources, span, tree, scope_i)
}

fn duplicate_term(tree: &mut Tree,
                  scope_i: Index,
                  orig_i: Index,
                  orig_span: Span)
                  -> Index {
    if let Some(id) = node::get_ident(tree, orig_i).cloned() {
        let dup_i = node::ident(id, tree);
        tree[dup_i].set_span(orig_span);
        dup_i
    } else {
        let mut replace_link = tree.replacement_link(orig_i).unwrap();
        let id = format!("$rel_common_{}", orig_i);
        let def_i = node::definition(id.clone(), scope_i, orig_i, tree);
        tree[def_i].set_span(orig_span);
        Link::child(scope_i).link(def_i, tree);
        let ident_i = node::resolved_ident(def_i, tree);
        tree[ident_i].set_span(orig_span);
        replace_link.link(ident_i, tree);
        let dup_i = node::resolved_ident(def_i, tree);
        tree[dup_i].set_span(orig_span);
        dup_i
    }
}

fn parse_expr_relational(sources: &SourceSet,
                         span: &mut Span,
                         tree: &mut Tree,
                         scope_i: Index)
                         -> Result<Index, Error> {
    let mut lhs_i = parse_expr_bitwise_cat(sources, span, tree, scope_i)?;
    let mut peek = *span;
    let mut token = Token::next(sources, &mut peek)?;
    match token.kind {
        TokenKind::IsEquals | TokenKind::NotEquals | TokenKind::Less |
        TokenKind::LessEquals | TokenKind::Greater |
        TokenKind::GreaterEquals => *span = peek,
        _ => return Ok(lhs_i),
    }
    let mut rhs_i = parse_expr_bitwise_cat(sources, span, tree, scope_i)?;
    let mut rhs_span = tree[rhs_i].span();
    let mut rel_i = node::expr_relational(token.kind, lhs_i, rhs_i, tree);
    let rel_span = tree[lhs_i].span() + rhs_span;
    tree[rel_i].set_span(rel_span);
    let mut top_i = rel_i;
    let mut and_link: Option<Link> = None;
    loop {
        peek = *span;
        token = Token::next(sources, &mut peek)?;
        match token.kind {
            TokenKind::IsEquals | TokenKind::NotEquals | TokenKind::Less |
            TokenKind::LessEquals | TokenKind::Greater |
            TokenKind::GreaterEquals => *span = peek,
            _ => return Ok(top_i),
        }
        // move common expression
        let dup_i = duplicate_term(tree, scope_i, rhs_i, rhs_span);
        let prev_rel_i = rel_i;
        lhs_i = dup_i;
        rhs_i = parse_expr_bitwise_cat(sources, span, tree, scope_i)?;
        rel_i = node::expr_relational(token.kind, lhs_i, rhs_i, tree);
        tree[rel_i].set_span(rhs_span);
        let and_i = node::expr_bitwise(TokenKind::And, prev_rel_i, rel_i, tree);
        if let Some(ref mut link) = and_link {
            link.link(and_i, tree);
        }
        and_link = Some(Link::sibling(prev_rel_i));
        let and_span = tree[prev_rel_i].span();
        tree[and_i].set_span(and_span);
        if prev_rel_i == top_i {
            top_i = and_i;
        }
        let mut parent_i = rel_i;
        loop {
            rhs_span += tree[parent_i].span();
            tree[parent_i].set_span(rhs_span);
            if parent_i == top_i {
                break;
            }
            parent_i = tree.parent(parent_i).unwrap();
        }
    }
}

fn parse_expr_bitwise_cat(sources: &SourceSet,
                          span: &mut Span,
                          tree: &mut Tree,
                          scope_i: Index)
                          -> Result<Index, Error> {
    let mut first: Option<TokenKind> = None;
    let mut lhs_i = parse_expr_shift(sources, span, tree, scope_i)?;
    loop {
        let mut peek = *span;
        let token = Token::next(sources, &mut peek)?;
        match token.kind {
            TokenKind::And | TokenKind::Or | TokenKind::Xor |
            TokenKind::Nand | TokenKind::Nor | TokenKind::Xnor |
            TokenKind::Cat => *span = peek,
            _ => return Ok(lhs_i),
        }
        let mixage = match (first, token.kind) {
            (None, _) => {
                first = Some(token.kind);
                None
            }
            (Some(first), second) if first == second => None,
            (Some(TokenKind::Cat), _) => {
                Some("concatenation with bitwise operators")
            }
            (_, TokenKind::Cat) => Some("bitwise with concatenation operators"),
            _ => Some("different bitwise operators"),
        };
        if let Some(mixage) = mixage {
            return Err(token.span.err(sources,
                                      &format!("cannot mix {}, consider \
                                                using brackets",
                                               mixage)));
        }
        let rhs_i = parse_expr_shift(sources, span, tree, scope_i)?;
        let expr_i = match token.kind {
            TokenKind::Cat => node::expr_concatenation(lhs_i, rhs_i, tree),
            _ => node::expr_bitwise(token.kind, lhs_i, rhs_i, tree),
        };
        let expr_span = tree[lhs_i].span() + tree[rhs_i].span();
        tree[expr_i].set_span(expr_span);
        lhs_i = expr_i;
    }
}

fn parse_expr_shift(sources: &SourceSet,
                    span: &mut Span,
                    tree: &mut Tree,
                    scope_i: Index)
                    -> Result<Index, Error> {
    let mut lhs_i = parse_expr_add(sources, span, tree, scope_i)?;
    loop {
        let mut peek = *span;
        let token = Token::next(sources, &mut peek)?;
        match token.kind {
            TokenKind::ShiftLeft | TokenKind::ShiftRight => *span = peek,
            _ => return Ok(lhs_i),
        }
        let rhs_i = parse_expr_add(sources, span, tree, scope_i)?;
        let shift_i = node::expr_shift(token.kind, lhs_i, rhs_i, tree);
        let shift_span = tree[lhs_i].span() + tree[rhs_i].span();
        tree[shift_i].set_span(shift_span);
        lhs_i = shift_i;
    }
}

fn parse_expr_add(sources: &SourceSet,
                  span: &mut Span,
                  tree: &mut Tree,
                  scope_i: Index)
                  -> Result<Index, Error> {
    let mut lhs_i = parse_expr_mul(sources, span, tree, scope_i)?;
    loop {
        let mut peek = *span;
        let token = Token::next(sources, &mut peek)?;
        match token.kind {
            TokenKind::Plus | TokenKind::Minus => *span = peek,
            _ => return Ok(lhs_i),
        };
        let rhs_i = parse_expr_mul(sources, span, tree, scope_i)?;
        let add_i = node::expr_addition(token.kind, lhs_i, rhs_i, tree);
        let add_span = tree[lhs_i].span() + tree[rhs_i].span();
        tree[add_i].set_span(add_span);
        lhs_i = add_i;
    }
}

fn parse_expr_mul(sources: &SourceSet,
                  span: &mut Span,
                  tree: &mut Tree,
                  scope_i: Index)
                  -> Result<Index, Error> {
    let mut lhs_i = parse_expr_pow(sources, span, tree, scope_i)?;
    loop {
        let mut peek = *span;
        let token = Token::next(sources, &mut peek)?;
        match token.kind {
            TokenKind::Times | TokenKind::Division | TokenKind::Remainder => {
                *span = peek
            }
            _ => return Ok(lhs_i),
        }
        let rhs_i = parse_expr_pow(sources, span, tree, scope_i)?;
        let mul_i = node::expr_multiplication(token.kind, lhs_i, rhs_i, tree);
        let mul_span = tree[lhs_i].span() + tree[rhs_i].span();
        tree[mul_i].set_span(mul_span);
        lhs_i = mul_i;
    }
}

fn parse_expr_pow(sources: &SourceSet,
                  span: &mut Span,
                  tree: &mut Tree,
                  scope_i: Index)
                  -> Result<Index, Error> {
    let mut lhs_i = parse_expr_cast(sources, span, tree, scope_i)?;
    loop {
        let mut peek = *span;
        if Token::next(sources, &mut peek)?.kind != TokenKind::Power {
            return Ok(lhs_i);
        }
        *span = peek;
        let rhs_i = parse_expr_cast(sources, span, tree, scope_i)?;
        let pow_i = node::expr_power(lhs_i, rhs_i, tree);
        let pow_span = tree[lhs_i].span() + tree[rhs_i].span();
        tree[pow_i].set_span(pow_span);
        lhs_i = pow_i;
    }
}

fn parse_expr_cast(sources: &SourceSet,
                   span: &mut Span,
                   tree: &mut Tree,
                   scope_i: Index)
                   -> Result<Index, Error> {
    let mut lhs_i = parse_expr_unary(sources, span, tree, scope_i)?;
    loop {
        let mut peek = *span;
        let token = Token::next(sources, &mut peek)?;
        match token.kind {
            TokenKind::KeywordAs => *span = peek,
            _ => return Ok(lhs_i),
        };
        let rhs_i = parse_expr_unary(sources, span, tree, scope_i)?;
        let cast_i = node::expr_cast(lhs_i, rhs_i, tree);
        let cast_span = tree[lhs_i].span() + tree[rhs_i].span();
        tree[cast_i].set_span(cast_span);
        lhs_i = cast_i;
    }
}

fn parse_expr_unary(sources: &SourceSet,
                    span: &mut Span,
                    tree: &mut Tree,
                    scope_i: Index)
                    -> Result<Index, Error> {
    let token = match parse_token_unary(sources, span) {
        Some(token) => token,
        _ => return parse_expr_brackets(sources, span, tree, scope_i),
    };
    let rhs_i = parse_expr_unary(sources, span, tree, scope_i)?;
    let unary_i = node::expr_unary(token.kind, rhs_i, tree);
    let unary_span = token.span + tree[rhs_i].span();
    tree[unary_i].set_span(unary_span);
    Ok(unary_i)
}

fn parse_expr_brackets(sources: &SourceSet,
                       span: &mut Span,
                       tree: &mut Tree,
                       scope_i: Index)
                       -> Result<Index, Error> {
    let mut peek = *span;
    let token = Token::next(sources, &mut peek)?;
    let is_range = match token.kind {
        TokenKind::OpenParens => false,
        TokenKind::OpenSquare => true,
        _ => return parse_expr_term(sources, span, tree, scope_i),
    };
    let brackets_i = if !is_range {
        *span = peek;
        let inner_i = parse_expr(sources, span, tree, scope_i)?;
        parse_token(sources, span, TokenKind::CloseParens, "expected ')'")?;
        inner_i
    } else {
        parse_range(sources, span, tree, scope_i)?
    };
    parse_accessing(sources, span, tree, brackets_i, scope_i)
}

fn parse_expr_term(sources: &SourceSet,
                   span: &mut Span,
                   tree: &mut Tree,
                   scope_i: Index)
                   -> Result<Index, Error> {
    let token = Token::next(sources, span)?;
    let term_i = match token.kind {
        TokenKind::Point => {
            let word_token = Token::next(sources, span)?;
            if word_token.kind != TokenKind::Word {
                return Err(word_token.span.err(sources, "expected identifier"));
            }
            let i = node::outer_ident(word_token.span.get(sources).to_string(),
                                      tree);
            tree[i].set_span(token.span + word_token.span);
            i
        }
        TokenKind::Word => {
            let i = node::ident(token.span.get(sources).to_string(), tree);
            tree[i].set_span(token.span);
            i
        }
        TokenKind::ZeroX | TokenKind::ZeroZ | TokenKind::Integer |
        TokenKind::IntegerXz => {
            let (val, mode) = match token.kind {
                TokenKind::ZeroX => (0.into(), (-1).into()),
                TokenKind::ZeroZ => ((-1).into(), (-1).into()),
                TokenKind::Integer | TokenKind::IntegerXz => {
                    token.integer_literal(sources)
                }
                _ => unreachable!(),
            };
            let tv_i = node::typed_value_integer(val, mode, tree);
            set_tv_span(tree, tv_i, token.span);
            tv_i
        }
        TokenKind::Float => {
            let val = token.float_literal(sources);
            let tv_i = node::typed_value_float(val, tree);
            set_tv_span(tree, tv_i, token.span);
            tv_i
        }
        _ => return Err(token.span.err(sources, "expected expression")),
    };
    parse_accessing(sources, span, tree, term_i, scope_i)
}

fn parse_expr_field(sources: &SourceSet,
                    span: &mut Span,
                    tree: &mut Tree,
                    expr_i: Index)
                    -> Result<Index, Error> {
    parse_token(sources, span, TokenKind::Point, "expected '.'")?;
    let token = Token::next(sources, span)?;
    let ident = match token.kind {
        TokenKind::Word => token.span.get(sources).to_string(),
        _ => return Err(token.span.err(sources, "expected field identifier")),
    };
    let field_i = node::field(ident, None, tree);
    tree[field_i].set_span(token.span);
    let access_i = node::expr_field(expr_i, field_i, tree);
    let access_span = tree[expr_i].span() + token.span;
    tree[access_i].set_span(access_span);
    Ok(access_i)
}

fn parse_expr_param(sources: &SourceSet,
                    span: &mut Span,
                    tree: &mut Tree,
                    expr_i: Index,
                    scope_i: Index)
                    -> Result<Index, Error> {
    parse_token(sources, span, TokenKind::OpenParens, "expected '('")?;
    let params_i = node::expr_parameters(expr_i, tree);
    let mut need_comma = false;
    let mut link = Link::sibling(expr_i);
    loop {
        let mut peek = *span;
        let token = Token::next(sources, &mut peek)?;
        match token.kind {
            TokenKind::Comma if need_comma => {
                *span = peek;
                need_comma = false;
                continue;
            }
            TokenKind::CloseParens => {
                *span = peek;
                let params_span = tree[expr_i].span() + token.span;
                tree[params_i].set_span(params_span);
                return Ok(params_i);
            }
            _ => {}
        }
        if need_comma {
            return Err(token.span.err(sources, "expected ',' or ')'"));
        }
        let param_i = parse_expr(sources, span, tree, scope_i)?;
        link.link(param_i, tree);
        need_comma = true;
    }
}

fn parse_expr_index(sources: &SourceSet,
                    span: &mut Span,
                    tree: &mut Tree,
                    expr_i: Index,
                    scope_i: Index)
                    -> Result<Index, Error> {
    let range_i = parse_range(sources, span, tree, scope_i)?;
    let access_i = node::expr_index(expr_i, range_i, tree);
    let access_span = tree[expr_i].span() + tree[range_i].span();
    tree[access_i].set_span(access_span);
    Ok(access_i)
}

fn parse_range(sources: &SourceSet,
               span: &mut Span,
               tree: &mut Tree,
               scope_i: Index)
               -> Result<Index, Error> {
    let open =
        parse_token(sources, span, TokenKind::OpenSquare, "expected '['")?;
    let mut peek = *span;
    let lhs_i = match Token::next(sources, &mut peek)?.kind {
        TokenKind::Colon | TokenKind::PlusColon | TokenKind::MinusColon => {
            node::typed_value_integer(0.into(), 0.into(), tree)
        }
        _ => parse_expr(sources, span, tree, scope_i)?,
    };
    let token = Token::next(sources, span)?;
    let is_sized = match token.kind {
        TokenKind::Colon => false,
        TokenKind::PlusColon | TokenKind::MinusColon => true,
        TokenKind::CloseSquare => {
            let range_single_i = node::range_single(lhs_i, tree);
            tree[range_single_i].set_span(open.span + token.span);
            return Ok(range_single_i);
        }
        _ => return Err(token.span.err(sources, "expected ':', '+:' or '-:'")),
    };
    let is_to_end = !is_sized &&
                    {
        peek = *span;
        parse_token(sources, &mut peek, TokenKind::CloseSquare, "expected ']'")
            .is_ok()
    };
    let range_i = if is_to_end {
        node::range_to_end(lhs_i, tree)
    } else {
        let rhs_i = parse_expr(sources, span, tree, scope_i)?;
        if is_sized {
            let is_minus = token.kind == TokenKind::MinusColon;
            node::range_sized(is_minus, lhs_i, rhs_i, tree)
        } else {
            node::range(lhs_i, rhs_i, tree)
        }
    };
    let close =
        parse_token(sources, span, TokenKind::CloseSquare, "expected ']'")?;
    tree[range_i].set_span(open.span + close.span);
    Ok(range_i)
}

fn parse_accessing(sources: &SourceSet,
                   span: &mut Span,
                   tree: &mut Tree,
                   mut expr_i: Index,
                   scope_i: Index)
                   -> Result<Index, Error> {
    loop {
        let mut peek = *span;
        let token = Token::next(sources, &mut peek)?;
        match token.kind {
            TokenKind::Point => {
                expr_i = parse_expr_field(sources, span, tree, expr_i)?
            }
            TokenKind::OpenParens => {
                expr_i = parse_expr_param(sources, span, tree, expr_i, scope_i)?
            }
            TokenKind::OpenSquare => {
                expr_i = parse_expr_index(sources, span, tree, expr_i, scope_i)?
            }
            _ => return Ok(expr_i),
        }
    }
}

fn parse_token(sources: &SourceSet,
               span: &mut Span,
               required: TokenKind,
               expected: &str)
               -> Result<Token, Error> {
    let token = Token::next(sources, span)?;
    match token.kind {
        x if x == required => Ok(token),
        _ => Err(token.span.err(sources, expected)),
    }
}

fn parse_token_unary(sources: &SourceSet, span: &mut Span) -> Option<Token> {
    let mut peek = *span;
    let token = match Token::next(sources, &mut peek) {
        Ok(s) => s,
        Err(_) => return None,
    };
    match token.kind {
        TokenKind::Not | TokenKind::And | TokenKind::Or | TokenKind::Xor |
        TokenKind::Plus | TokenKind::Minus => {
            *span = peek;
            Some(token)
        }
        TokenKind::Nand | TokenKind::Nor | TokenKind::Xnor => {
            // we cannot return !& !| or !^, we need to return just the !
            span.skip_whitespace(sources);
            let mut ret = Token {
                kind: TokenKind::Not,
                span: *span,
            };
            let len = match span.get(sources).char_indices().next() {
                Some((index, c)) if c == '!' => index + c.len_utf8(),
                _ => unreachable!(),
            };
            ret.span.set_len(len as u32);
            span.advance_start(len as u32);
            Some(ret)
        }
        _ => None,
    }
}

fn set_tv_span(tree: &mut Tree, tv_i: Index, span: Span) {
    tree[tv_i].set_span(span);
    let (ty_i, val_i) = node::get_typed_value(tree, tv_i).unwrap();
    tree[ty_i].set_span(span);
    tree[val_i].set_span(span);
}
